package GrapherGUI;


import java.lang.Math;
import java.awt.*;
import java.awt.geom.*;
import java.awt.event.*;
import javax.swing.event.*;
import javax.swing.*;


/** Creates a panel object which draws the given functions.
 * @author Bo Shi
 */
public class Graph extends JPanel {

    /** width in pixels of the graph
     */    
    public final int X = 600;

    /** height in pixels of the visible graph
     */        
    public final int Y = 500;

    /** constant passed into zoom
     * @see Graph.zoom
     */        
    public final short ZOOM_IN_X = 0;

    /** constant passed into zoom
     * @see Graph.zoom
     */        
    public final short ZOOM_IN_Y = 1;

    /** constant passed into zoom
     * @see Graph.zoom
     */        
    public final short ZOOM_OUT_X = 2;

    /** constant passed into zoom
     * @see Graph.zoom
     */        
    public final short ZOOM_OUT_Y = 3;        

    /** maximum number of functions one plot can display
     */        
    public final int MAXFUNCTIONS = 10;
    int _numf = 0;      // number of functions
    int _density = 100;

    double _tickIncrementX = 1.0;
    double _tickIncrementY = 1.0;
    int _tickDensityX = 4;
    int _tickDensityY = 4;
    
    boolean _show_axes_number_labels = true;
	
    double _minx, _maxx;
    double _miny, _maxy;
    double _xscale, _yscale;
    double _mousex = 0.0;
    double _mousey = 0.0;
    Dimension _D = new Dimension(X, Y);

    /** pixel position of the origin on the plot
     */        
    Point _origin;

    /** List of functions associated with this graph.
     */        
    Function[] _f = new Function[MAXFUNCTIONS];

    /** Object containing parameter information.
     */        
    SliderMatrix _sm;

    /** displays the location of the mouse in cartesian coordinates
     */        
    JLabel _mouseLabel = new JLabel("Mouse @ (0, 0)");
    boolean _do_linkPoints = true;
    boolean _do_showKey = true;

    ImageIcon _xicon = null;
    ImageIcon _yicon = null;
    String _xaxisLabel, _yaxisLabel;
        
    Image _image;
    Graphics _g;
    FontMetrics _fm;

    /**
     * Class constructor.  Sets x values (-10, 10) and y values (-10, 10).
     */
    public Graph() {
        setPreferredSize(_D);
        addMouseMotionListener(new MouseTracker());
        _fm = getFontMetrics(getFont());
        _minx = -10.0;
        _maxx = 10.0;
        _miny = -10.0;
        _maxy = 10.0;
        _xscale = (double) X / Math.abs(_maxx - _minx);
        _yscale = (double) Y / Math.abs(_maxy - _miny);
        _origin = new Point(X / 2, Y / 2);
        _sm = new SliderMatrix(5);
        updateTickDensity();
    }

    public void showLabels(boolean b) {
        _show_axes_number_labels = b;
    }

    /** Set the frequency in which tick marks and labels show up
     * on the x-axis..
     *
     * @param d value between tick marks
     */
    public void setTickIncrementX(double d) {
        _tickIncrementX = d;
        updateTickDensity();
    }

    /** Set the frequency in which tick marks and labels show up
     * on the y-axis..
     *
     * @param d value between tick marks.
     */
    public void setTickIncrementY(double d) {
        _tickIncrementY = d;
        updateTickDensity();
    }

    public void setTickDensityX(int i) {
        _tickDensityX = i;
    }

    public void setTickDensityY(int i) {
        _tickDensityY = i;
    }

    public int getTickDensityX() {
        return _tickDensityX;
    }

    public int getTickDensityY() {
        return _tickDensityY;
    }

    public void updateTickDensity() {
        _tickDensityX = (int) (Math.abs(_maxx - _minx) / _tickIncrementX);
        _tickDensityY = (int) (Math.abs(_maxy - _miny) / _tickIncrementY);
    }

    public void updateTickIncrement() {
        _tickIncrementX = Math.abs(_maxx - _minx) / (double) _tickDensityX;
        _tickIncrementY = Math.abs(_maxy - _miny) / (double) _tickDensityY;
    }

    /** Class constructor specifying dimensions of graph.
     *
     * @param sm the list of parameters
     * @param minx minimum x value
     * @param maxx maximum x value
     * @param miny minimum y value
     * @param maxy maximum y value
     */
    public Graph(SliderMatrix sm, double minx, double maxx, double miny, double maxy) {
        setPreferredSize(_D);
        addMouseMotionListener(new MouseTracker());
        _minx = minx;
        _maxx = maxx;
        _miny = miny;
        _maxy = maxy;
        scale();
        _sm = sm;
    }
        
    /**
     * Called after a change in scale.  This resets appropriate variables
     * so that paint() works correctly after a zoom call.
     */
    public void scale() {
        _xscale = (double) X / Math.abs(_maxx - _minx);
        _yscale = (double) Y / Math.abs(_maxy - _miny);
        // need to recalc origin _origin = new Point(X / 2, Y / 2);

        /* _origin = new Point((int) Math.abs(_minx * _xscale),
         (int) Math.abs(_maxy * _yscale)); */
        _origin = new Point((int) (-_minx * _xscale), (int) (_maxy * _yscale));
        repaint();
    }
        
    /** Get the number of points that are used to draw each function.
     * @return density
     */        
    public int getDensity() {
        return _density;
    }
        
    /** Set the number of points that are used to draw each function.
     * @param d number of points to use in drawing each function
     * @return true if set was successful otherwise false
     */        
    public boolean setDensity(int d) {
        if (d > 0 && d <= X) {
            _density = d;
            return true;
        }
        return false;
    }
            
    /**
     * Adds a function to the graph.  A total of ten functions is allowed on
     * any one plot. 
     *
     * @param  f    function to be added to the plot
     * @return      true if adding was successful.  False if f is null or if
     *              the plot has ten functions already
     * @see         Function
     */
    public boolean addFunc(Function f) {
        if (_numf >= MAXFUNCTIONS || f == null) {
            return false;
        }
        _f[_numf] = f;
        _numf++;
        repaint();
        return true;
    }
        
    /** Set an icon for the axes of the graph.
     *
     * @param xfile filename of the independent icon file (must be JPEG or GIF)
     * @param yfile filename of the dependent icon (must be JPEG or GIF)
     *
     */
    public void setAxesIcons(String xfile, String yfile) {
        _xicon = Util.createImageIcon("/" + xfile, "Independent Variable");
        _yicon = Util.createImageIcon("/" + yfile, "Dependent Variable");
    }
        
    /** Set an string for the axes of the graph.  The strings will be used in the case
     * that icons are not provided or the images cannot be found.
     * @param x x-axis label
     * @param y y-axis label
     */        
    public void setAxisLabels(String x, String y) {
        _xaxisLabel = x;
        _yaxisLabel = y;
    }
        
    /** Change the state of the boolean handling whether points on the graph are tied
     * together by lines.
     */        
    public void changeStateLinkPoints() {
        _do_linkPoints = (_do_linkPoints ? false : true);
        repaint();
    }

    public void changeStateShowLabels() {
        _show_axes_number_labels = (_show_axes_number_labels ? false : true);
        repaint();
    }
        
    /** Show/remove key.
     */        
    public void changeStateShowKey() {
        _do_showKey = (_do_showKey ? false : true);
        repaint();
    }

    /** Increases/decreases the domain and range of the graph according to
     * the parameter it recieves.
     *
     * @param how  either zoom in x/y or out x/y
     * @see Graph.ZOOM_IN_X
     * @see Graph.ZOOM_IN_Y
     * @see Graph.ZOOM_OUT_X
     * @see Graph.ZOOM_OUT_Y
     */
    public void zoom(short how) {
        double domain = Math.abs(_maxx - _minx);
        double range = Math.abs(_maxy - _miny);
        double newdomain = domain;
        double newrange = range;

        if (how == ZOOM_IN_X) {
            newdomain = domain * 0.5;
        } else if (how == ZOOM_IN_Y) {
            newrange = range * 0.5;
        } else if (how == ZOOM_OUT_X) {
            newdomain = domain * 2.0;
        } else if (how == ZOOM_OUT_Y) {
            newrange = range * 2.0;
        }

        if (_miny > 0) {
            _maxy = _miny + newrange;
        }
        if (_maxy < 0) {
            _miny = _maxy - newrange;
        }
        if (_minx > 0) {
            _maxx = _minx + newdomain;
        }
        if (_maxx < 0) {
            _minx = _maxx - newdomain;
        }
        if (_minx <= 0 && _maxx >= 0) {
            _minx = _minx / domain * newdomain;
            _maxx = newdomain + _minx;
        }
        if (_miny <= 0 && _maxy >= 0) {
            _miny = _miny / range * newrange;
            _maxy = newrange + _miny;
        }

        scale();
        repaint();
    }
                
    /*** DRAWING FUNCTIONS ***/
    
    /*************************/

    /** Redraws the graph according to the current parameter set.
     *
     * @param g    graphics object on which to draw
     * @see Graphics
     * @see Function
     */
    public void paint(Graphics g) {
        if (_image == null) {
            _image = createImage((int) _D.getWidth(), (int) _D.getHeight());
            _g = _image.getGraphics();
            _fm = _g.getFontMetrics();
        }
        _fm = _g.getFontMetrics();

        _g.setColor(Color.white);
        _g.fillRect(0, 0, _D.width, _D.height);

        drawGrid();
        drawFunctions();
        drawKey();

        g.drawImage(_image, 0, 0, null);
    }

    /**
     * Draws the axes and all appropriate labels.
     * <p>
     * If the user specified an image to use for a variable name, that will
     * be drawn near its respective axis.  Otherwise, the string name of
     * the parameter will be drawn.
     */
    public void drawGrid() {
        Graphics2D g2 = (Graphics2D) _g;

        g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                RenderingHints.VALUE_ANTIALIAS_ON);

        FontMetrics fm = g2.getFontMetrics();

        // draw axes
        g2.setPaint(Color.red);
        g2.draw(new Line2D.Double(_origin.getX(), 0.0, _origin.getX(),
                (double) Y));
        g2.draw(new Line2D.Double(0.0, _origin.getY(), (double) X,
                _origin.getY()));

        // draw a black border around the plot surface
        g2.setPaint(Color.black);
        g2.draw(new Line2D.Double(0.0, 0.0, 0.0, (double) Y));
        g2.draw(new Line2D.Double(0.0, 0.0, (double) X, 0.0));
        g2.draw(new Line2D.Double((double) X - 1.0, (double) Y - 1.0,
                (double) X - 1.0, 0.0));
        g2.draw(new Line2D.Double((double) X - 1.0, (double) Y - 1.0, 0.0,
                (double) Y - 1.0));
        g2.setPaint(Color.red);

        // calculate tick increment and then draw
        updateTickIncrement();

        // draw ticks and labels
        double d = _tickIncrementX;

        g2.setPaint(Color.red);
        // System.out.println("increments (x,y) : (" + _tickIncrementX + ", " +
        // _tickIncrementY + ")");
        // draw numbers and ticks
        while (d < _maxx) {
            Point p = cart2screen(d, 0.0);

            drawXTickLabel(Util.formatDouble(d), (int) p.getX(), (int) p.getY());
            d += _tickIncrementX;
        }
        d = -_tickIncrementX;
        while (d > _minx) {
            Point p = cart2screen(d, 0.0);

            drawXTickLabel(Util.formatDouble(d), (int) p.getX(), (int) p.getY());
            d -= _tickIncrementX;
        }
        d = _tickIncrementY;
        while (d < _maxy) {
            Point p = cart2screen(0.0, d);

            drawYTickLabel(Util.formatDouble(d), (int) p.getX(), (int) p.getY());
            d += _tickIncrementY;
        }
        d = -_tickIncrementY;                
        while (d > _miny) {
            Point p = cart2screen(0.0, d);

            drawYTickLabel(Util.formatDouble(d), (int) p.getX(), (int) p.getY());
            d -= _tickIncrementY;
        }
                
        // draw the axis labels
                
        if (_xicon != null) {
            Image icon = _xicon.getImage();

            _g.drawImage(icon, X - icon.getWidth(this) - 1,
                    (int) _origin.getY() + 1, this);
        } else if (_xaxisLabel != null) {
            Image icon = _xicon.getImage();

            _g.drawString(_xaxisLabel, X - icon.getWidth(this) - 1,
                    (int) _origin.getY() + 1);
        }
        if (_yicon != null) {
            Image icon = _yicon.getImage();

            _g.drawImage(icon, (int) _origin.getX() + 1,
                    Y - icon.getHeight(this) - 1, this);
        } else if (_yaxisLabel != null) {
            Image icon = _yicon.getImage();

            _g.drawString(_yaxisLabel, (int) _origin.getX() + 1,
                    Y - icon.getHeight(this) - 1);
        }
    }

    private void drawXTickLabel(String str, int xcoord, int ycoord) {
        // takes the SCREEN coordinates where the label should be drawn
        Graphics2D g2 = (Graphics2D) _g;

        g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                RenderingHints.VALUE_ANTIALIAS_ON);
        FontMetrics fm = g2.getFontMetrics();

        g2.setPaint(Color.red);

        int yoffset = -3;

        if (ycoord > Y) {
            yoffset *= 1;
            ycoord = Y;
        }
        if (ycoord < 0) {
            yoffset *= -1;
            ycoord = 0;
        }

        g2.draw(new Line2D.Double(xcoord, ycoord, xcoord, ycoord + yoffset));
        g2.setPaint(Color.black);
        if (true == _show_axes_number_labels) {
            xcoord -= fm.stringWidth(str) / 2;
            yoffset = (int) ((double) yoffset * 1.4);
            if (ycoord < (fm.getAscent() + 2)) {
                yoffset += fm.getAscent() + Math.abs(yoffset) * 2;
            }
            g2.drawString(str, xcoord, ycoord + yoffset);
        }
    }

    private void drawYTickLabel(String str, int xcoord, int ycoord) {
        // takes the SCREEN coordinates where the label should be drawn
        Graphics2D g2 = (Graphics2D) _g;

        g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                RenderingHints.VALUE_ANTIALIAS_ON);
        FontMetrics fm = g2.getFontMetrics();

        g2.setPaint(Color.red);
        int xoffset = 3;
        int h = fm.getAscent();
        int w = fm.stringWidth(str);

        if (xcoord > X) {
            xoffset *= -1;
            xcoord = X;
        }
        if (xcoord < 0) {
            xoffset *= 1;
            xcoord = 0;
        }

        g2.draw(new Line2D.Double(xcoord, ycoord, xcoord + xoffset, ycoord));
        g2.setPaint(Color.black);
        if (true == _show_axes_number_labels) {
            xoffset = (int) ((double) xoffset * 1.4);
            if (xcoord > (X - w - xoffset - 4)) {
                xoffset -= (w + Math.abs(xoffset) * 2);
            }
            g2.drawString(str, xcoord + xoffset, ycoord + h / 2);
        }

    }

    /**
     * Draws each function.
     */
    public void drawFunctions() {
        Graphics2D g2 = (Graphics2D) _g;

        // for each function in the function list, draw the graph
        for (int i = 0; i < _numf; i++) {
            g2.setPaint(_f[i].color());

            if (_f[i].type() == Function.FCNTYPE_SINGLEVAL) {
                df_oneval(_f[i]);
            } else {
                df_onepass(_f[i]);
            }
        }
    }

    /** Function drawer.  This does it in one pass using an existing
     * array of vlues
     */
    private void df_onepass(Function fcn) {
        Graphics2D g2 = (Graphics2D) _g;

        g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                RenderingHints.VALUE_ANTIALIAS_ON);

        double[] X, Y;	// these should not be initialized
        Point p;
        double prevx = _minx;
        double prevy = _miny;

        X = new double[_density];
        Y = new double[_density];
        fcn.farray(_minx, _maxx, _density, X, Y, _sm);

        for (int i = 0; i < X.length; i++) {
            p = cart2screen(X[i], Y[i]);
            if (i != 0 && _do_linkPoints == true) {
                Point end = cart2screen(prevx, prevy);

                if (inLimits(end) || inLimits(p)) {
                    g2.draw(new Line2D.Double(end.getX(), end.getY(), p.getX(),
                            p.getY()));
                }
            }
			
            if (!_do_linkPoints && inLimits(p)) {
                g2.draw(new Ellipse2D.Double(p.getX(), p.getY(), 1.0, 1.0));
            }

            prevx = X[i];
            prevy = Y[i];
        }

    }

    private void df_oneval(Function fcn) {
        Graphics2D g2 = (Graphics2D) _g;

        g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                RenderingHints.VALUE_ANTIALIAS_ON);

        double incr = Math.abs(_maxx - _minx) / (double) _density;
        double x, y;
        double prevx = _minx;
        double prevy = _miny;

        x = _minx;

        while (x < _maxx) {
            y = fcn.f(x, _sm);
            Point p = cart2screen(x, y);

            if (x != _minx && _do_linkPoints == true) {
                // connect the lines
                Point end = cart2screen(prevx, prevy);

                if (inLimits(end) || inLimits(p)) {
                    g2.draw(new Line2D.Double(end.getX(), end.getY(), p.getX(),
                            p.getY()));
                }
							 
            }

            if (!_do_linkPoints && inLimits(p)) {
                g2.draw(new Ellipse2D.Double(p.getX(), p.getY(), 1.0, 1.0));
            }
            prevx = x;
            prevy = y;
            x += incr;
        }
    }

    /** Checks to make sure a pixel coordinate is in bounds.
     * @param p the pixel coordinates to test
     * @return true if the pixel is on the drawing surface
     */
    private boolean inLimits(Point p) {
        if (((int) p.getX() < X && (int) p.getX() >= 0)
                || ((int) p.getY() < Y && (int) p.getY() >= 0)) {
            return true;
        }
        return false;
    }
        
    /** Draw the key onto the top left corner of the graph.  It will be drawn
     * transparent if this feature is available.
     */        
    public void drawKey() {
        if (!_do_showKey) {
            return;
        }
        int i;
        Graphics2D g2 = (Graphics2D) _g;

        g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING,
                RenderingHints.VALUE_ANTIALIAS_ON);
        String[] txt = new String[_numf];

        for (i = 0; i < _numf; i++) {
            txt[i] = _f[i].name();
        }
                
        // set width and height
        int h = _fm.getHeight() * (_numf + 1);
        int a = _fm.getAscent();
        int w = -1;

        for (i = 0; i < _numf; i++) {
            if (_fm.stringWidth(txt[i]) > w) {
                w = _fm.stringWidth(txt[i]);
            }
        }
        w += 50;
        // draw
        g2.setPaint(new Color(55, 55, 55, 130));
        g2.fill(new Rectangle2D.Double(5.0, 5.0, (double) w, (double) h));
        for (i = 0; i < _numf; i++) {
            g2.setPaint(_f[i].color());
            int y = 10 + a * (i + 1);

            g2.draw(new Line2D.Double(10.0, (double) (y - a / 2), 30.0,
                    (double) (y - a / 2)));
            g2.drawString(txt[i], 40, y);
        }
		
    }

    /** Creates an arrow polygon.
     *
     * @param x     x coordinate (pixels) of the center of the arrow
     * @param y     y coordinate (pixels) of the center of the arrow
     * @param theta angle (in polar coordinates) at which arrow points
     * @return      polygon which represents the three vertices of the
     *              arrow aligned so that it points at angle theta.
     * @see getTheta
     */
    private Polygon createArrow(int x, int y, double theta) {
        // the three points which define the triangle polygon that represents
        // the arrow are defined in polar coordinates on the circle of radius
        // 'r'.  Here, we rotate all points theta degrees, then convert them
        // to cartesian coordinates.  After this, the polygon is translated to
        // the coordinates (x, y) as specified by the input.
        double r = 3.0;
        double angle1 = 0.0 + theta;
        double angle2 = 2.0 * Math.PI / 3.0 + theta;
        double angle3 = 2.0 * Math.PI * 2.0 / 3.0 + theta;
        Polygon arrow = new Polygon();

        arrow.addPoint(x + (int) (r * Math.cos(angle1)),
                y + (int) (r * Math.sin(angle1)));
        arrow.addPoint(x + (int) (r * Math.cos(angle2)),
                y + (int) (r * Math.sin(angle2)));
        arrow.addPoint(x + (int) (r * Math.cos(angle3)),
                y + (int) (r * Math.sin(angle3)));
        return arrow;
    }

    /**
     * Converts a point on the cartesian coordinate plane to a pixel
     * coordinate on the drawing canvas. 
     *
     * @param  x    x value (cartesian coordinates)
     * @param  y    y value (cartesian coordinates)
     * @return      Point object representing the coordinate of the pixel
     *              at which (x, y) represent.
     * @see         Point
     */
    public Point cart2screen(double x, double y) {
        return (new Point((int) _origin.getX() + (int) (_xscale * x),
                (int) _origin.getY() - (int) (_yscale * y)));
    }
        
    /**
     * Converts a point on the canvas plane to a point on the cartesian
     * coordinate plane.
     *
     * @param  x    pixel x coordinate
     * @param  y    pixel y coordinate
     * @return      Point object representing the coordinate on the
     *              cartesian plane.
     * @see         Point2D
     */
    public Point2D.Double screen2cart(int x, int y) {
        // double xpixels = _origin.getX() + (double) x;
        // double ypixels = _origin.getY() - (double) y;

        double xpixels = (double) x - _origin.getX();
        double ypixels = (double) (-y) + _origin.getY();

        return (new Point2D.Double(xpixels / _xscale, ypixels / _yscale));
    }

    /**
     * Gets the angle in polar coordinates of a line.  The first endpoint
     * represents the origin. 
     *
     * @param  l    object representing the line
     * @return      angle in radians
     * @see         Line2D
     */
    public double getTheta(Line2D l) {
        double theta;

        theta = Math.atan((l.getY2() - l.getY1()) / (l.getX2() - l.getX1()));
        if (l.getX1() > l.getX2()) {
            theta = theta + Math.PI;
        }

        return theta;
    }
        
    /**
     * Returns a string representation of the current mouse coordinates
     * on the graph.
     *
     * @return      string in the form of (x,y)
     */
    public String strMouseCoord() {
        return ("(" + Util.formatDouble(_mousex) + ", "
                + Util.formatDouble(_mousey) + ")");
    }
        
    /**
     * Returns a link to the mouse label.  Used in GUI class.
     *
     * @return      JLabel showing mouse coordinates.
     */
    public JLabel mouseLabel() {
        return _mouseLabel;
    }
   
    /** MouseMotionListener object to handle mouse events.
     * @see MouseMotionListener
     */        
    public class MouseTracker implements MouseMotionListener {

        /** Keep track of the mouse's position in cartesian coordinates and report back
         * to the label displaying this information.
         * @param e object containing cursor position information
         */            
        public void mouseMoved(MouseEvent e) {
            Point2D.Double p = screen2cart(e.getX(), e.getY());

            _mousex = p.getX();
            _mousey = p.getY();
            _mouseLabel.setText("Mouse at " + strMouseCoord());
        }

        /** No behavior is defined for this event.
         * @param e not used
         */                
        public void mouseDragged(MouseEvent e) {}
    }

    /** Get the minimum x value
     * @return minimum x
     */
    public double minx() {
        return _minx;
    }

    /** Get the minimum y value
     * @return minimum y
     */
    public double miny() {
        return _miny;
    }

    /** Get the maximum x value
     * @return maximum x
     */
    public double maxx() {
        return _maxx;
    }

    /** Get the maximum y value
     * @return maximum y
     */
    public double maxy() {
        return _maxy;
    }

    public void setDomain(double min, double max) {
        if (min < max) {
            _minx = min;
            _maxx = max;
        }
        scale();
    }

    public void setRange(double min, double max) {
        if (min < max) {
            _miny = min;
            _maxy = max;
        }
        scale();
    }

}
