package GrapherGUI;


import java.awt.*;


/** The Graph class maintains a list of Function objects
 * which need to be plotted.  The user must define the
 * function which takes one independent value and a list
 * of possible parameter values.  This function then
 * returns the dependent value.  To see how to grab
 * parameter values, check the SliderMatrix class.
 * @author Bo Shi
 */
public abstract class Function {

    /** Make the <b>Graph</b> class use <b>f()</b> when plotting.
     */
    public static final short FCNTYPE_SINGLEVAL = 1;

    /** Make the <b>Graph</b> class use <b>farray()</b> when plotting.
     */
    public static final short FCNTYPE_ARRAY = 2;

    /** What kind of algorithm is used?
     */

    /** The GrapherGUI class requires information about the algorithm
     * type used to perform calculations.  If the algorithm for your
     * mathematical function is fairly simple, implementing the
     * function using <b>f()</b> is probably the simplest solution.
     * Alternatively, if your algorithm calculates a whole vector
     * of points, you may use <b>farray()</b> to interface with
     * the GUI.
     */
    public abstract short type();

    /** A Graph object uses this user defined object to calculate the
     * set of coordinates used to draw the function.  Alternatively,
     * the other method <b>farray()</b> can be used to populate two
     * arrays with the independent and dependent variables.
     * @param in a value of the independent variable
     * @param p containing the parameter values.
     * @see SliderMatrix
     * @return value of the dependent variable associated with the input
     */    
    public abstract double f(double in, SliderMatrix p);

    public abstract void farray(
            double minx,
            double maxx,
            int density,
            double[] varIndep,
            double[] varDep,
            SliderMatrix p);
        
    /** Defines the default color of the function.  Should be overridden.
     * If the subclass does not define this function, the color will default
     * to blue.
     * @return Color object
     */        
    public Color color() {
        return (new Color(0, 0, 255, 153));
    }
        
    /** The name of the function to be displayed in the key.
     * @return a description of the function
     */        
    public abstract String name();
}
