package GrapherGUI;


import java.awt.*;
import java.util.*;
import javax.swing.event.*;
import javax.swing.*;


/** Each ParameterPanel is a self contained object that maintains the information
 * about one parameter in the environment.  It does all event handling relating
 * to the slider and the value label. A screenshot is provided below:<br>
 * <img src="ParameterPanel.png">
 */
public class ParameterPanel extends JPanel {
    ImageIcon _imageIcon;
    String _paramString;
    JSlider _slider;
    JTextField _textField;
    JLabel _label;
    int _numTicks;

    /** JSliders have an integer range.  We must set some resolution in integers
     * and then convert that value to a double that represents the correct value
     * as specified by the range given.
     * @see _convFactor
     */        
    final int _resolution = 2000;
    double _convFactor;		// needed to give the users
    double _min, _max, _value;

    /** Construtor requires initialization information.
     * @param min minimum value this parameter can take
     * @param max maximum value this parameter can take
     * @param start initial value this parameter takes
     * @param numticks number of tickmarks on the slider
     * @param imagefile an picture of the parameter name.  null is a valid input.
     * @param name string identifier for this parameter
     */        
    public ParameterPanel(double min, double max, double start, int numticks, String imagefile, String name) {
        _paramString = name;
        if (imagefile == null) {
            // System.out.println("Image set to NULL");
            _imageIcon = null;
        } else {
            // System.out.println("Image set to [" + imagefile + "]");
            _imageIcon = Util.createImageIcon("/" + imagefile, name);
        }
        _min = min;
        _max = max;
        _value = start;
        _numTicks = numticks;
        _convFactor = (double) _resolution / Math.abs(_max - _min);

        // set slider
        Hashtable slabels = new Hashtable();

        slabels.put(new Integer(0), new JLabel(Util.formatDouble(_min)));
        slabels.put(new Integer(_resolution),
                new JLabel(Util.formatDouble(_max)));
        _slider = new JSlider(0, _resolution, val2slider(_value));
        _slider.setLabelTable(slabels);
        _slider.setMinorTickSpacing(_resolution / _numTicks);
        _slider.setPaintTicks(true);
        _slider.setPaintLabels(true);
        _slider.addChangeListener(new sliderListener());

        // set text field
        _textField = new JTextField(Util.formatDouble(_value), 6);
        _textField.setEditable(false);

        if (_imageIcon == null) {
            // System.out.println("Should be adding paramString...");
            _label = new JLabel(_paramString);
        } else {
            _label = new JLabel(_imageIcon);
        }

        // set tooltip fields
        _slider.setToolTipText("Change the value of " + _paramString);
        _textField.setToolTipText("Current value of " + _paramString);
        setBorder(BorderFactory.createLineBorder(Color.black));
        add(_label);
        add(_slider);
        add(_textField);
    }

    public void resize(double min, double max) {
        if (min > max) {
            return;
        }

        _min = min;
        _max = max;
        _convFactor = (double) _resolution / Math.abs(_max - _min);

        Hashtable slabels = new Hashtable();

        slabels.put(new Integer(0), new JLabel(Util.formatDouble(_min)));
        slabels.put(new Integer(_resolution),
                new JLabel(Util.formatDouble(_max)));
        _slider.setLabelTable(slabels);
    }
		
    /** Translates slider position to actual parameter value.
     * @param sliderval current slider position
     * @return double value
     */        
    private double slider2val(int sliderval) {
        return _min + (double) sliderval / _convFactor;
    }

    /** Translates a value to a slider position.
     * @param value a parameter value
     * @return integer value
     */        
    private int val2slider(double value) {
        if (value < _min || value > _max) {
            return 500;
        }
        return (int) ((value - _min) * _convFactor);
    }

    /** Get the current value of the parameter.
     * @return parameter value
     */        
    public double val() {
        // report current value
        return slider2val(_slider.getValue());
    }

    /** Returns the JSlider object contained in this class.
     * @return JSlider object
     */        
    public JSlider slider() {
        return _slider;
    }

    /** increments the slider by <b>num</b>.  This is used by the
     * animation methods.  Returns true unless the slider has wrapped
     * around because the slider has reached maximum.
     */
    public boolean incr(double num) {
        boolean r = true;
        int max = _slider.getMaximum();
        int val = _slider.getValue();
        int dx = (int) (num * _convFactor);

        if ((val + dx) > max) {
            // System.out.println("FALSE val " + val + ", dx " + dx + ", max "
            // + max);
            _slider.setValue(val + dx);
            return false;
        }

        // System.out.println("TRUE val " + val + ", dx " + dx + ", max " + max);
        _slider.setValue(val + dx);
        return true;
    }

    /** Reset the slider to its minimum value.  Used by animation methods.
     */
    public void snapLeft() {
        _slider.setValue(1);
    }
        
    /** Get the string identifier of this parameter.
     * @return name
     */        
    public String name() {
        return _paramString;
    }

    /** Event handler for changes in slider position.  This updates the JTextField
     * displaying the current parameter value.
     */        
    class sliderListener implements ChangeListener {

        /** Show what the current value is in a JTextField.
         * @param e not used
         */            
        public void stateChanged(ChangeEvent e) {
            _textField.setText(Util.formatDouble(slider2val(_slider.getValue())));
        }
    }

    public double min() {
        return _min;
    }

    public double max() {
        return _max;
    }
}
